(* Avant tout, un tableau contenant les résultats attendus au problème des 
  n reines. Source : https://www.durangobill.com/N_Queens.html *)
let solutions = 
  [|0; 1; 0; 0; 2; 10; 4; 40; 92; 352; 724; 2680; 14200; 73712; 365596|]



(** Question 1. **)

(* conflit: int -> int -> int -> int -> bool
   Renvoie true si les reines placées en (x1, y1) et (x2, y2) sont en conflit *)
let conflit x1 y1 x2 y2 =
  x1 = x2 || y1 = y2 || abs (x2-x1) = abs (y2-y1)



(** Question 2. **)

(* est_solution: int array -> bool
   À partir de la configuration tab, on vérifie pour chaque paire de reines 
   distinctes si un conflit existe. On renvoie true ssi il n'y en a aucun. *)
let est_solution tab =
  let n = Array.length tab in
  let res = ref true in
  for i = 0 to n-1 do
    for j = i+1 to n-1 do
      if conflit i tab.(i) j tab.(j) then
        res := false
    done  
  done;
  !res

  

(** Question 3. **)

(* succ: int array -> bool
   Fonction intermédiaire.
   Cette fonction prend en paramètre une configuration, et la modifie en place
   vers la configuation suivante. Appeller cette fonction de façon répétée 
   permet d'énumérer toutes les configuation. Par convention, pour une taille n
   donnée, la première configuration est [|0; 0; ... ; 0|] et la dernière est
   [| n-1; n-1; ... ; n-1 |]. La fonction renvoie false si on lui donne en 
   paramètre la dernière configuration, et true sinon. *)
let succ tab =
  let n = Array.length tab in 
  (* aux_succ : int -> bool 
     Fonction auxiliaire récursive qui incrémente le i-ème chiffre de tab.*)
  let rec aux_succ i =
    (* Si i=n, c'est qu'on vient de passe tous les chiffres de n-1 à 0. On
       renvoie false pour indiquer que c'était la dernière configuration. *)
    if i = n then false
    else begin
      tab.(i) <- (tab.(i) + 1) mod n;
      (* Si le i-ème chiffre passe de n-1 à 0, on passe au chiffre suivant *)
      if tab.(i) = 0 then
        aux_succ (i+1)
      else
        true
    end
  in 
    aux_succ 0


(* Quelques tests pour s'assurer que la fonction succ marche bien, mais aussi
   pour bien comprendre son fonctionnement. *)
let () =
  let t1 = [|0; 2; 1|] in
  assert (succ t1 = true);
  assert (t1 = [|1; 2; 1|]);
  assert (succ t1 = true);
  assert (t1 = [|2; 2; 1|]);
  assert (succ t1 = true);
  assert (t1 = [|0; 0; 2|]);
  assert (succ t1 = true);
  assert (t1 = [|1; 0; 2|]);

  let t2 = [|3; 3; 3; 3|] in
  assert (succ t2 = false); (* <- false ici ! *)
  assert (t2 = [|0; 0; 0; 0|]);
  assert (succ t2 = true);
  assert (t2 = [|1; 0; 0; 0|]);
  
  Printf.printf("Tous les tests sont passés !\n")

(* nb_solutions: int -> int
   Renvoie le nombre de solution au problème des n reines.
   Avec la fonction succ, le principe est le suivant : on énumère toutes les
   configuration depuis [|0; 0; ... 0|] avec la fonction succ. Quand celle-ci
   renvoie false, on s'arrête. On compte le nombre de solution n'entraînant pas
   de conflit. *)
let nb_solutions n =
  let tab = Array.make n 0 in
  let compteur = ref 0 in
  let continuer = ref true in
  while !continuer do
    if est_solution tab then
      incr compteur;
      continuer := succ tab;
  done;
  !compteur

(* On teste que la fonction nb_solution donne bien les valeurs attendues (qui
   sont dans le tableau solution). n=8 est le plus grand entier pour lequel le 
   programme termine rapidement. Pour n=9, cela pendrait plusieurs minutes. *)
let () =
  for n = 1 to 8 do
    assert (nb_solutions n = solutions.(n));
  done;
  Printf.printf("Tous les tests sont passés !\n")



(** Question 5. **)

(* est_valide: int -> int array -> bool
   Similaire à la fonction est_solution. À partir de la configuration tab, on 
   vérifie pour chaque paire de reines distinctes des k premières lignes si un 
   conflit existe. *)
let est_valide k tab = 
  let res = ref true in
  for i = 0 to k-1 do
    for j = i+1 to k-1 do
      if conflit i tab.(i) j tab.(j) then
        res := false
    done  
  done;
  !res



(** Question 6. **)

(* nb_solutions2: int -> int
   Renvoie le nombre de solution au problème des n reines.
   On utilise cette fois le principe du retour sur trace. *)
let nb_solutions2 n =
  let tab = Array.make n 0 in

  (* aux_nb_solutions2: int -> int
     Fonction auxiliaire récursive, revoyant le nombre de solution au problème 
     des n reines, pour lequel les k première lignes sont déjà remplies. *)
  let rec aux_nb_solutions2 k =
    (* Si toutes les lignes sont fixées et qu'on arrive jusqu'ici, on a trouvé
       1 solution. *)
    if k = n then 1
    else begin
      (* Sinon, pour la ligne courante d'indice k, on teste toutes les positions
         possibles. Pour celle qui sont valides, on appelle récursivement la
         fonction pour la ligne suivante, et on fait la somme des nombres de
         solutions trouvées. *)
      let count = ref 0 in
      for i = 0 to n-1 do
        tab.(k) <- i;
        if est_valide (k+1) tab then
          count := !count + aux_nb_solutions2 (k+1)
      done;
      !count
    end
  in aux_nb_solutions2 0
      

let () =
  (* On peut monter jusqu'à 12 en un temps court avec le retour sur trace. *)
  for n = 1 to 12 do
    assert (nb_solutions2 n = solutions.(n));
  done;
  Printf.printf("Tous les tests sont passés !\n")
